// Copyright (c) 1999 Philip A. Hardin (pahardin@cs.utexas.edu)
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License v2 or later.


#ifndef EDGE_h
#define EDGE_h


#include "general.h"



/************************************************************************/
struct edge
{ int	beginPtNum,		// point number that starts this edge
	endPtNum;		// point number that ends this edge
};


#endif
