// Copyright (c) 1999 Philip A. Hardin (pahardin@cs.utexas.edu)
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License v2 or later.


#ifndef _RENDERGL_H
#define _RENDERGL_H
#include "general.h"
#include "gfxtarget.h"
#include "xregion3d.h"

#ifndef NO_OPENGL
#include <GL/gl.h>
#endif


//---------------------------------------------------------------------------
void DrawGL(const xpanel3d& panel, const pt3d viewPts[], const ulong colors[],
            const gfxTarget& gt);

//---------------------------------------------------------------------------
void DrawGL(xregion3d& r, const pt3d viewPts[], const gfxTarget& gt,
            const pt3d& viewPos, const pt2d& clip, bool solid, const ulong colors[]);
#endif

