// Copyright (c) 1999 Philip A. Hardin (pahardin@cs.utexas.edu)
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License v2 or later.


#include "bbcolor.h"

char *bbColorNames[totalColors]= {
  "black","white", "gray10","gray20","gray30","gray40","gray50","gray60",
  "gray70","gray75","gray80","gray85","gray90","gray95",
  "saddle brown", "sienna", "brown", "firebrick", "red",
  "orange red", "dark orange", "orange", "gold", "yellow",
  "green yellow", "green", "lawn green", "lime green", "green3",
  "forest green", "dark green", "dark olive green", "olive drab",
  "cyan", "light sky blue", "deep sky blue", "dodger blue", "blue",
  "steelblue1", "steelblue2", "steelblue3", "steel blue", "steelblue4",
  "slategray1", "slategray2", "slategray3", "slate gray", "slategray4",
  "navy", "violet", "turquoise", "gray50", "gray30"
};
