// Copyright (c) 1999 Philip A. Hardin (pahardin@cs.utexas.edu)
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License v2 or later.


#ifndef BB_h
#define BB_h


#include "general.h"
#include "shdregion3d.h"


// Convenient type aliases:
typedef shdRegion3d rgn3;
typedef region2d  rgn2;
typedef rgn3::coord coord;


#endif
